@@ -105,7 +105,7 @@ const owl_cmd commands_to_init[]
 
   OWLCMD_ARGS("zwrite", owl_command_zwrite, OWL_CTX_INTERACTIVE,
 	      "send a zephyr",
-	      "zwrite [-n] [-C] [-c class] [-i instance] [-r realm] [-O opcde] [<user> ...] [-m <message...>]",
+	      "zwrite [-n] [-C] [-c class] [-i instance] [-r realm] [-O opcode] [<user> ...] [-m <message...>]",
 	      "Zwrite send a zephyr to the one or more users specified.\n\n"
 	      "The following options are available:\n\n"
 	      "-m    Specifies a message to send without prompting.\n"
@@ -135,9 +135,9 @@ const owl_cmd commands_to_init[]
 	      "loopwrite",
 	      "Send a local message.\n"),
 
-  OWLCMD_ARGS("zcrypt", owl_command_zcrypt, OWL_CTX_INTERACTIVE,
+  OWLCMD_ARGS("zcrypt", owl_command_zwrite, OWL_CTX_INTERACTIVE,
 	      "send an encrypted zephyr",
-	      "zcrypt [-n] [-C] [-c class] [-i instance] [-r realm] [-O opcde] [-m <message...>]\n",
+	      "zcrypt [-n] [-C] [-c class] [-i instance] [-r realm] [-O opcode] [-m <message...>]\n",
 	      "Behaves like zwrite but uses encryption.  Not for use with\n"
 	      "personal messages\n"),
   
@@ -147,7 +147,7 @@ const owl_cmd commands_to_init[]
 	      "If -e is specified, the zwrite command line is presented to\n"
 	      "allow editing.\n\n"
 	      "If 'sender' is specified, reply to the sender.\n\n"
-	      "If 'all' or no args are specified, reply publically to the\n"
+	      "If 'all' or no args are specified, reply publicly to the\n"
 	      "same class/instance for non-personal messages and to the\n"
 	      "sender for personal messages.\n\n"
 	      "If 'zaway' is specified, replies with a zaway message.\n\n"),
@@ -265,7 +265,7 @@ const owl_cmd commands_to_init[]
 	      "suppress a given zephyr triplet",
 	      "zpunt <class> <instance> [recipient]\n"
 	      "zpunt <instance>",
-	      "The zpunt command will supress message to the specified\n"
+	      "The zpunt command will suppress messages to the specified\n"
 	      "zephyr triplet.  In the second usage messages are suppressed\n"
 	      "for class MESSAGE and the named instance.\n\n"
 	      "SEE ALSO:  zunpunt, show zpunts\n"),
@@ -282,7 +282,7 @@ const owl_cmd commands_to_init[]
 	      "suppress an arbitrary filter",
 	      "punt <filter-text>",
 	      "punt <filter-text (multiple words)>\n"
-	      "The punt command will supress message to the specified\n"
+	      "The punt command will suppress messages to the specified\n"
 	      "filter\n\n"
 	      "SEE ALSO:  unpunt, zpunt, show zpunts\n"),
 
@@ -585,7 +585,7 @@ const owl_cmd commands_to_init[]
 	      "by the\n 'view_home' variable.  The style can be set by listing the\n"
               "name style after the -s argument.\n"
 	      "\n"
-	      "The other usages listed above are abbreivated forms that simply set\n"
+	      "The other usages listed above are abbreviated forms that simply set\n"
 	      "the filter of the current view. The -d option allows you to write a\n"
               "filter expression that will be dynamically created by owl and then\n"
               "applied as the view's filter\n"
@@ -593,7 +593,7 @@ const owl_cmd commands_to_init[]
 
   OWLCMD_ARGS("smartnarrow", owl_command_smartnarrow, OWL_CTX_INTERACTIVE,
 	      "view only messages similar to the current message",
-	      "smartnarrow [-i | --instance]  [-r | --relatde]",
+	      "smartnarrow [-i | --instance]  [-r | --related]",
 	      "If the curmsg is a personal message narrow\n"
 	      "   to the conversation with that user.\n"
 	      "If the curmsg is a <MESSAGE, foo, *>\n"
@@ -613,7 +613,7 @@ const owl_cmd commands_to_init[]
 	      "If the curmsg is a <MESSAGE, foo, *>\n"
 	      "   message, the filter is to that instance.\n"
 	      "If the curmsg is a class message, the filter is that class.\n"
-	      "If the curmsg is a class message and '-i' is specied\n"
+	      "If the curmsg is a class message and '-i' is specified\n"
 	      "    the filter is to that class and instance.\n"),
 
   OWLCMD_ARGS("viewclass", owl_command_viewclass, OWL_CTX_INTERACTIVE,
@@ -653,6 +653,7 @@ const owl_cmd commands_to_init[]
 	      "show styles\n"
 	      "show subscriptions / show subs\n"
 	      "show terminal\n"
+	      "show timers\n"
 	      "show variables\n"
 	      "show variable <variable>\n"
 	      "show version\n"
@@ -673,7 +674,7 @@ const owl_cmd commands_to_init[]
 	      "Show styles will list the names of all styles available\n"
 	      "for formatting messages.\n\n"
 	      "Show variables will list the names of all variables.\n\n"
-	      "Show errors will show a list of errors ecountered by Owl.\n\n"
+	      "Show errors will show a list of errors encountered by Owl.\n\n"
 	      "SEE ALSO: filter, view, alias, bindkey, help\n"),
   
   OWLCMD_ARGS("delete", owl_command_delete, OWL_CTX_INTERACTIVE,
@@ -742,7 +743,7 @@ const owl_cmd commands_to_init[]
   OWLCMD_ARGS("setsearch", owl_command_setsearch, OWL_CTX_INTERACTIVE,
 	      "set the search highlight string without searching",
 	      "setsearch <string>",
-	      "The setsearch command highlights all occurences of its\n"
+	      "The setsearch command highlights all occurrences of its\n"
           "argument and makes it the default argument for future\n"
           "search commands, but does not move the cursor.  With\n"
           "no argument, it makes search highlighting inactive."),
@@ -767,6 +768,16 @@ const owl_cmd commands_to_init[]
               "message <message>",
               ""),
 
+  OWLCMD_ARGS("add-cmd-history", owl_command_add_cmd_history, OWL_CTX_ANY,
+              "Add a command to the history",
+              "add-cmd-history <cmd>",
+              ""),
+
+  OWLCMD_ARGS("with-history", owl_command_with_history, OWL_CTX_ANY,
+              "Run a command and store it into the history",
+              "with-history <cmd>",
+              ""),
+
   OWLCMD_VOID("yes", owl_command_yes, OWL_CTX_RECV,
               "Answer yes to a question",
               "yes",
@@ -998,7 +1009,29 @@ const owl_cmd commands_to_init[]
 		  "exits the popless window",
 		  "", ""),
 
-  OWLCMD_ALIAS("webzephyr", "zwrite daemon.webzephyr -c webzephyr -i"),
+  OWLCMD_ARGS_CTX("popless:start-command", owl_viewwin_start_command,
+		  OWL_CTX_POPLESS,
+		  "starts a command line in the popless",
+		  "popless:start-command [initial-value]",
+		  "Initializes the command field to initial-value"),
+
+  OWLCMD_ARGS_CTX("popless:search", owl_viewwin_command_search, OWL_CTX_POPLESS,
+		  "search lines for a particular string",
+		  "popless:search [-r] [<string>]",
+		  "The popless:search command will find lines that contain the\n"
+		  "specified string and scroll the popwin there.  If no string\n"
+		  "argument is supplied then the previous one is used.  By\n"
+		  "default searches are done forwards; if -r is used the search\n"
+		  "is performed backwards"),
+
+  OWLCMD_ARGS_CTX("popless:start-search", owl_viewwin_command_start_search, OWL_CTX_POPLESS,
+		  "starts a command line to search for particular string",
+		  "popless:start-search [-r] [inital-value]",
+		  "Initializes the command-line to search for initial-value. If\n"
+		  "-r is used, the search will be performed backwards.\n\n"
+                  "SEE ALSO: popless:search"),
+
+  OWLCMD_ALIAS("webzephyr", "zwrite " OWL_WEBZEPHYR_PRINCIPAL " -c " OWL_WEBZEPHYR_CLASS " -i"),
 
   /* This line MUST be last! */
   { NULL, NULL, NULL, NULL, 0, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL }
@@ -1296,14 +1329,14 @@ char *owl_command_smartnarrow(int argc, const char *const *argv, const char *buf
   int instance = 0, related = 0, i;
   const char **tmp_argv = owl_malloc(sizeof(char *) * argc);
 
-  for (i = 0; i < argc; i++)
-    tmp_argv[i] = argv[i];
-
   static const struct option options[] = {
     {"instance", 0, 0, 'i'},
     {"related",  0, 0, 'r'},
     {NULL,       0, 0, 0}};
 
+  for (i = 0; i < argc; i++)
+    tmp_argv[i] = argv[i];
+
   optind = 0;
   while ((opt = getopt_long(argc, (char **)tmp_argv, "ir", options, NULL)) != -1) {
     switch (opt) {
@@ -1370,7 +1403,6 @@ void owl_command_resize(void)
 void owl_command_redisplay(void)
 {
   owl_function_full_redisplay();
-  owl_global_set_needrefresh(&g);
 }
 
 char *owl_command_get_shift(int argc, const char *const *argv, const char *buff)
@@ -1386,8 +1418,6 @@ char *owl_command_get_shift(int argc, const char *const *argv, const char *buff)
 void owl_command_set_shift(int shift)
 {
   owl_global_set_rightshift(&g, shift);
-  owl_mainwin_redisplay(owl_global_get_mainwin(&g));
-  owl_global_set_needrefresh(&g);
 }
 
 void owl_command_unsuball(void)
@@ -1607,32 +1637,32 @@ char *owl_command_print(int argc, const char *const *argv, const char *buff)
 
 char *owl_command_exec(int argc, const char *const *argv, const char *buff)
 {
-  return owl_function_exec(argc, argv, buff, 0);
+  return owl_function_exec(argc, argv, buff, OWL_OUTPUT_RETURN);
 }
 
 char *owl_command_pexec(int argc, const char *const *argv, const char *buff)
 {
-  return owl_function_exec(argc, argv, buff, 1);
+  return owl_function_exec(argc, argv, buff, OWL_OUTPUT_POPUP);
 }
 
 char *owl_command_aexec(int argc, const char *const *argv, const char *buff)
 {
-  return owl_function_exec(argc, argv, buff, 2);
+  return owl_function_exec(argc, argv, buff, OWL_OUTPUT_ADMINMSG);
 }
 
 char *owl_command_perl(int argc, const char *const *argv, const char *buff)
 {
-  return owl_function_perl(argc, argv, buff, 0);
+  return owl_function_perl(argc, argv, buff, OWL_OUTPUT_RETURN);
 }
 
 char *owl_command_pperl(int argc, const char *const *argv, const char *buff)
 {
-  return owl_function_perl(argc, argv, buff, 1);
+  return owl_function_perl(argc, argv, buff, OWL_OUTPUT_POPUP);
 }
 
 char *owl_command_aperl(int argc, const char *const *argv, const char *buff)
 {
-  return owl_function_perl(argc, argv, buff, 2);
+  return owl_function_perl(argc, argv, buff, OWL_OUTPUT_ADMINMSG);
 }
 
 char *owl_command_multi(int argc, const char *const *argv, const char *buff)
@@ -1913,25 +1943,30 @@ void owl_command_status(void)
 
 char *owl_command_zwrite(int argc, const char *const *argv, const char *buff)
 {
-  owl_zwrite z;
+  owl_zwrite *z;
 
   if (!owl_global_is_havezephyr(&g)) {
     owl_function_makemsg("Zephyr is not available");
     return(NULL);
   }
   /* check for a zwrite -m */
-  owl_zwrite_create_from_line(&z, buff);
-  if (owl_zwrite_is_message_set(&z)) {
-    owl_function_zwrite(buff, NULL);
-    owl_zwrite_cleanup(&z);
-    return (NULL);
+  z = owl_zwrite_new(buff);
+  if (!z) {
+    owl_function_error("Error in zwrite arguments");
+    return NULL;
+  }
+
+  if (owl_zwrite_is_message_set(z)) {
+    owl_function_zwrite(z, NULL);
+    owl_zwrite_delete(z);
+    return NULL;
   }
-  owl_zwrite_cleanup(&z);
 
   if (argc < 2) {
+    owl_zwrite_delete(z);
     owl_function_makemsg("Not enough arguments to the zwrite command.");
   } else {
-    owl_function_zwrite_setup(buff);
+    owl_function_zwrite_setup(z);
   }
   return(NULL);
 }
@@ -2021,31 +2056,6 @@ char *owl_command_loopwrite(int argc, const char *const *argv, const char *buff)
   return(NULL);
 }
 
-char *owl_command_zcrypt(int argc, const char *const *argv, const char *buff)
-{
-  owl_zwrite z;
-
-  if (!owl_global_is_havezephyr(&g)) {
-    owl_function_makemsg("Zephyr is not available");
-    return(NULL);
-  }
-  /* check for a zcrypt -m */
-  owl_zwrite_create_from_line(&z, buff);
-  if (owl_zwrite_is_message_set(&z)) {
-    owl_function_zcrypt(buff, NULL);
-    owl_zwrite_cleanup(&z);
-    return (NULL);
-  }
-  owl_zwrite_cleanup(&z);
-
-  if (argc < 2) {
-    owl_function_makemsg("Not enough arguments to the zcrypt command.");
-  } else {
-    owl_function_zwrite_setup(buff);
-  }
-  return(NULL);
-}
-
 char *owl_command_reply(int argc, const char *const *argv, const char *buff)
 {
   int edit=0;
@@ -2183,10 +2193,6 @@ char *owl_command_view(int argc, const char *const *argv, const char *buff)
       owl_function_change_currentview_filter(owl_global_get_view_home(&g));
       argc--;
       argv++;
-    } else if (!strcmp(argv[0], "-r")) {
-      const char *foo;
-      foo=owl_function_create_negative_filter(owl_view_get_filtname(owl_global_get_current_view(&g)));
-      owl_function_change_currentview_filter(foo);
     } else if (!strcmp(argv[0], "-s")) {
       if (argc<2) {
 	owl_function_makemsg("Too few argments to the view command");
@@ -2248,6 +2254,8 @@ char *owl_command_show(int argc, const char *const *argv, const char *buff)
     owl_function_show_colors();
   } else if (!strcmp(argv[1], "styles")) {
     owl_function_show_styles();
+  } else if (!strcmp(argv[1], "timers")) {
+    owl_function_show_timers();
   } else if (!strcmp(argv[1], "subs") || !strcmp(argv[1], "subscriptions")) {
     owl_function_getsubs();
   } else if (!strcmp(argv[1], "terminal") || !strcmp(argv[1], "term")) {
@@ -2282,8 +2290,10 @@ char *owl_command_viewclass(int argc, const char *const *argv, const char *buff)
     return NULL;
   }
   filtname = owl_function_classinstfilt(argv[1], NULL, owl_global_is_narrow_related(&g));
-  owl_function_change_currentview_filter(filtname);
-  owl_free(filtname);
+  if (filtname) {
+    owl_function_change_currentview_filter(filtname);
+    owl_free(filtname);
+  }
   return NULL;
 }
 
@@ -2295,8 +2305,10 @@ char *owl_command_viewuser(int argc, const char *const *argv, const char *buff)
     return NULL;
   }
   filtname=owl_function_zuserfilt(argv[1]);
-  owl_function_change_currentview_filter(filtname);
-  owl_free(filtname);
+  if (filtname) {
+    owl_function_change_currentview_filter(filtname);
+    owl_free(filtname);
+  }
   return NULL;
 }
 
@@ -2386,7 +2398,7 @@ char *owl_command_colorview(int argc, const char *const *argv, const char *buff)
 
 char *owl_command_colorclass(int argc, const char *const *argv, const char *buff)
 {
-  const char *filtname;
+  char *filtname;
   
   if (argc < 3 || argc > 4) {
     owl_function_makemsg("Wrong number of arguments to colorclass command");
@@ -2394,7 +2406,10 @@ char *owl_command_colorclass(int argc, const char *const *argv, const char *buff
   }
 
   filtname=owl_function_classinstfilt(argv[1], NULL, owl_global_is_narrow_related(&g));
-  (void) owl_function_color_filter(filtname, argv[2], (argc == 4 ? argv[3] : NULL));
+  if (filtname) {
+    (void) owl_function_color_filter(filtname, argv[2], (argc == 4 ? argv[3] : NULL));
+    owl_free(filtname);
+  }
   return NULL;
 }
 
@@ -2558,9 +2573,11 @@ char *owl_command_search(int argc, const char *const *argv, const char *buff)
   }
     
   if (argc==1 || (argc==2 && !strcmp(argv[1], "-r"))) {
-    owl_function_search_continue(direction);
+    /* When continuing a search, don't consider the current message. */
+    owl_function_search_helper(false, direction);
   } else {
-    owl_function_search_start(buffstart, direction);
+    owl_function_set_search(buffstart);
+    owl_function_search_helper(true, direction);
   }
   
   return(NULL);
@@ -2571,7 +2588,7 @@ char *owl_command_setsearch(int argc, const char *const *argv, const char *buff)
   const char *buffstart;
 
   buffstart=skiptokens(buff, 1);
-  owl_function_search_start(*buffstart ? buffstart : NULL, OWL_DIRECTION_NONE);
+  owl_function_set_search(*buffstart ? buffstart : NULL);
   
   return(NULL);
 }
@@ -2632,6 +2649,46 @@ char *owl_command_message(int argc, const char *const *argv, const char *buff)
     return NULL;
 }
 
+char *owl_command_add_cmd_history(int argc, const char *const *argv, const char *buff)
+{
+  owl_history *hist;
+  const char *ptr;
+
+  if (argc < 2) {
+    owl_function_makemsg("usage: %s <commands> ...", argv[0]);
+    return NULL;
+  }
+
+  ptr = skiptokens(buff, 1);
+  hist = owl_global_get_cmd_history(&g);
+  owl_history_store(hist, ptr);
+  owl_history_reset(hist);
+  /* owl_function_makemsg("History '%s' stored successfully", ptr+1); */
+  return NULL;
+}
+
+char *owl_command_with_history(int argc, const char *const *argv, const char *buff)
+{
+  owl_history *hist;
+  const char *ptr;
+
+  if (argc < 2) {
+    owl_function_makemsg("usage: %s <commands> ...", argv[0]);
+    return NULL;
+  }
+
+  ptr = skiptokens(buff, 1);
+  if (!ptr) {
+    owl_function_makemsg("Parse error finding command");
+    return NULL;
+  }
+
+  hist = owl_global_get_cmd_history(&g);
+  owl_history_store(hist, ptr);
+  owl_history_reset(hist);
+  return owl_function_command(ptr);
+}
+
 void owl_command_yes(void)
 {
   owl_message *m;
@@ -2710,17 +2767,13 @@ void owl_command_edit_cancel(owl_editwin *e)
 
   owl_function_makemsg("Command cancelled.");
 
-  if(owl_editwin_get_echochar(e) == 0) {
-    hist=owl_editwin_get_history(e);
+  hist = owl_editwin_get_history(e);
+  if (hist) {
     owl_history_store(hist, owl_editwin_get_text(e));
     owl_history_reset(hist);
   }
 
-  owl_global_set_needrefresh(&g);
   owl_global_pop_context(&g);
-
-  owl_global_set_typwin_inactive(&g);
-  owl_editwin_delete(e);
 }
 
 void owl_command_edit_history_prev(owl_editwin *e)
@@ -2729,6 +2782,8 @@ void owl_command_edit_history_prev(owl_editwin *e)
   const char *ptr;
 
   hist=owl_editwin_get_history(e);
+  if (!hist)
+    return;
   if (!owl_history_is_touched(hist)) {
     owl_history_store(hist, owl_editwin_get_text(e));
     owl_history_set_partial(hist);
@@ -2737,8 +2792,6 @@ void owl_command_edit_history_prev(owl_editwin *e)
   if (ptr) {
     owl_editwin_clear(e);
     owl_editwin_insert_string(e, ptr);
-    owl_editwin_redisplay(e);
-    owl_global_set_needrefresh(&g);
   } else {
     owl_function_beep();
   }
@@ -2750,12 +2803,12 @@ void owl_command_edit_history_next(owl_editwin *e)
   const char *ptr;
 
   hist=owl_editwin_get_history(e);
+  if (!hist)
+    return;
   ptr=owl_history_get_next(hist);
   if (ptr) {
     owl_editwin_clear(e);
     owl_editwin_insert_string(e, ptr);
-    owl_editwin_redisplay(e);
-    owl_global_set_needrefresh(&g);
   } else {
     owl_function_beep();
   }
@@ -2765,8 +2818,6 @@ char *owl_command_edit_insert_text(owl_editwin *e, int argc, const char *const *
 {
   buff = skiptokens(buff, 1);
   owl_editwin_insert_string(e, buff);
-  owl_editwin_redisplay(e);
-  owl_global_set_needrefresh(&g);  
   return NULL;
 }
 
@@ -2779,12 +2830,15 @@ void owl_command_edit_done(owl_editwin *e)
     owl_history_reset(hist);
   }
 
-  owl_global_set_typwin_inactive(&g);
+  /* Take a reference to the editwin, so that it survives the pop
+   * context. TODO: We should perhaps refcount or otherwise protect
+   * the context so that, even if a command pops a context, the
+   * context itself will last until the command returns. */
+  owl_editwin_ref(e);
   owl_global_pop_context(&g);
-  owl_global_set_needrefresh(&g);
 
   owl_editwin_do_callback(e);
-  owl_editwin_delete(e);
+  owl_editwin_unref(e);
 }
 
 void owl_command_edit_done_or_delete(owl_editwin *e)
@@ -2803,8 +2857,14 @@ void owl_command_edit_done_or_delete(owl_editwin *e)
 
 void owl_command_popless_quit(owl_viewwin *vw)
 {
-  owl_popwin_close(owl_global_get_popwin(&g));
+  owl_popwin *pw;
+  pw = owl_global_get_popwin(&g);
+  owl_global_set_popwin(&g, NULL);
+  /* Kind of a hack, but you can only have one active viewwin right
+   * now anyway. */
+  if (vw == owl_global_get_viewwin(&g))
+    owl_global_set_viewwin(&g, NULL);
+  owl_viewwin_delete(vw);
+  owl_popwin_delete(pw);
   owl_global_pop_context(&g);
-  owl_viewwin_cleanup(vw);
-  owl_global_set_needrefresh(&g);
 }
