@@ -92,6 +92,36 @@ void owl_function_show_styles(void) {
   owl_fmtext_cleanup(&fm);
 }
 
+static void _owl_function_timer_append_fmtext(gpointer data, gpointer user_data) {
+  owl_fmtext *fm = user_data;
+  owl_timer *timer = data;
+  char *str = owl_sprintf("- %s: in %d seconds",
+                          timer->name ? timer->name : "(unnamed)",
+                          (int)(timer->time - time(NULL)));
+  owl_fmtext_append_normal(fm, str);
+  owl_free(str);
+  if (timer->interval) {
+    str = owl_sprintf(", repeat every %d seconds", timer->interval);
+    owl_fmtext_append_normal(fm, str);
+    owl_free(str);
+  }
+  owl_fmtext_append_normal(fm, "\n");
+}
+
+void owl_function_show_timers(void) {
+  owl_fmtext fm;
+  GList **timers;
+
+  owl_fmtext_init_null(&fm);
+  owl_fmtext_append_bold(&fm, "Active timers:\n");
+
+  timers = owl_global_get_timerlist(&g);
+  g_list_foreach(*timers, _owl_function_timer_append_fmtext, &fm);
+
+  owl_function_popless_fmtext(&fm);
+  owl_fmtext_cleanup(&fm);
+}
+
 char *owl_function_style_describe(const char *name) {
   const char *desc;
   char *s;
@@ -215,26 +245,19 @@ void owl_function_adminmsg(const char *header, const char *body)
 
   /* redisplay etc. */
   owl_mainwin_redisplay(owl_global_get_mainwin(&g));
-  owl_global_set_needrefresh(&g);
 }
 
 /* Create an outgoing zephyr message and return a pointer to it.  Does
  * not put it on the global queue, use owl_global_messagequeue_addmsg() for
  * that.
  */
-owl_message *owl_function_make_outgoing_zephyr(const char *body, const char *zwriteline, const char *zsig)
+owl_message *owl_function_make_outgoing_zephyr(const owl_zwrite *z)
 {
   owl_message *m;
-  owl_zwrite zw;
-
-  owl_zwrite_create_from_line(&zw, zwriteline);
-  owl_zwrite_set_zsig(&zw, zsig);
 
   /* create the message */
   m=owl_malloc(sizeof(owl_message));
-  
-  owl_message_create_from_zwrite(m, &zw, body);
-  owl_zwrite_cleanup(&zw);
+  owl_message_create_from_zwrite(m, z, owl_zwrite_get_message(z));
 
   return(m);
 }
@@ -280,6 +303,7 @@ owl_message *owl_function_make_outgoing_loopback(const char *body)
 void owl_function_start_edit_win(const char *line, void (*callback)(owl_editwin *), void *data, void (*cleanup)(void *))
 {
   owl_editwin *e;
+  owl_context *ctx;
   char *s;
 
   /* create and setup the editwin */
@@ -292,10 +316,13 @@ void owl_function_start_edit_win(const char *line, void (*callback)(owl_editwin
 
   owl_editwin_set_cbdata(e, data, cleanup);
   owl_editwin_set_callback(e, callback);
-  owl_global_push_context(&g, OWL_CTX_EDITMULTI, e, "editmulti");
+  ctx = owl_editcontext_new(OWL_CTX_EDITMULTI, e, "editmulti",
+                            owl_global_deactivate_editcontext, &g);
+  owl_global_push_context_obj(&g, ctx);
+
 }
 
-static void owl_function_write_setup(const char *line, const char *noun, void (*callback)(owl_editwin *))
+static void owl_function_write_setup(const char *noun)
 {
 
   if (!owl_global_get_lockout_ctrld(&g))
@@ -306,78 +333,75 @@ static void owl_function_write_setup(const char *line, const char *noun, void (*
     owl_function_makemsg("Type your %s below.  "
 			 "End with a dot on a line by itself.  ^C will quit.",
 			 noun);
-
-  owl_function_start_edit_win(line, callback,
-                              owl_strdup(line),
-                              owl_free);
 }
 
-void owl_function_zwrite_setup(const char *line)
+void owl_function_zwrite_setup(owl_zwrite *z)
 {
-  owl_zwrite z;
-  int ret;
-
-  /* check the arguments */
-  ret=owl_zwrite_create_from_line(&z, line);
-  if (ret) {
-    owl_function_error("Error in zwrite arguments");
-    owl_zwrite_cleanup(&z);
-    return;
-  }
-
   /* send a ping if necessary */
   if (owl_global_is_txping(&g)) {
-    owl_zwrite_send_ping(&z);
+    owl_zwrite_send_ping(z);
   }
-  owl_zwrite_cleanup(&z);
 
-  owl_function_write_setup(line, "zephyr", &owl_callback_zwrite);
+
+  owl_function_write_setup("zephyr");
+  owl_function_start_edit_win(z->zwriteline,
+                              &owl_callback_zwrite,
+                              z, (void(*)(void*))owl_zwrite_delete);
 }
 
 void owl_function_aimwrite_setup(const char *line)
 {
-  owl_function_write_setup(line, "message", &owl_callback_aimwrite);
+  owl_function_write_setup("message");
+  owl_function_start_edit_win(line,
+                              &owl_callback_aimwrite,
+                              owl_strdup(line),
+                              owl_free);
+
 }
 
 void owl_function_loopwrite_setup(void)
 {
-  owl_function_write_setup("loopwrite", "message", owl_callback_loopwrite);
+  owl_function_write_setup("message");
+  owl_function_start_edit_win("loopwrite",
+                              &owl_callback_loopwrite,
+                              NULL, NULL);
 }
 
 void owl_callback_zwrite(owl_editwin *e) {
-  char *command = owl_editwin_get_cbdata(e);
-  owl_function_zwrite(command,
-                      owl_editwin_get_text(e));
+  owl_zwrite *z = owl_editwin_get_cbdata(e);
+  owl_function_zwrite(z, owl_editwin_get_text(e));
 }
 
 /* send, log and display an outgoing zephyr.  If 'msg' is NULL
  * the message is expected to be set from the zwrite line itself
  */
-void owl_function_zwrite(const char *line, const char *msg)
+#ifdef HAVE_LIBZEPHYR
+void owl_function_zwrite(owl_zwrite *z, const char *msg)
 {
-  owl_zwrite z;
-  const char *mymsg;
   owl_message *m;
+  int ret;
 
-  if(!strncmp(line, "zcrypt", strlen("zcrypt"))) {
-    owl_function_zcrypt(line, msg);
+  if(strcmp(z->cmd, "zcrypt") == 0) {
+    owl_function_zcrypt(z, msg);
     return;
   }
 
   /* create the zwrite and send the message */
-  owl_zwrite_create_from_line(&z, line);
-  owl_zwrite_populate_zsig(&z);
+  owl_zwrite_populate_zsig(z);
   if (msg) {
-    owl_zwrite_set_message(&z, msg);
+    owl_zwrite_set_message(z, msg);
+  }
+  ret = owl_zwrite_send_message(z);
+  if (ret != 0) {
+    owl_function_makemsg("Error sending zephyr: %s", error_message(ret));
+    return;
   }
-  owl_zwrite_send_message(&z);
   owl_function_makemsg("Waiting for ack...");
 
   /* If it's personal */
-  if (owl_zwrite_is_personal(&z)) {
+  if (owl_zwrite_is_personal(z)) {
     /* create the outgoing message */
-    mymsg=owl_zwrite_get_message(&z);
-    m=owl_function_make_outgoing_zephyr(mymsg, line, owl_zwrite_get_zsig(&z));
+    m=owl_function_make_outgoing_zephyr(z);
 
     if (m) {
       owl_global_messagequeue_addmsg(&g, m);
@@ -385,63 +409,61 @@ void owl_function_zwrite(const char *line, const char *msg)
       owl_function_error("Could not create outgoing zephyr message");
     }
   }
-
-  /* free the zwrite */
-  owl_zwrite_cleanup(&z);
 }
+#else
+void owl_function_zwrite(owl_zwrite *z, const char *msg) {
+}
+#endif
 
 /* send, log and display an outgoing zcrypt zephyr.  If 'msg' is NULL
  * the message is expected to be set from the zwrite line itself
  */
-void owl_function_zcrypt(const char *line, const char *msg)
+void owl_function_zcrypt(owl_zwrite *z, const char *msg)
 {
-  owl_zwrite z;
-  const char *mymsg;
   char *cryptmsg;
   owl_message *m;
   const char *argv[7];
   char *zcrypt;
   int rv, status;
+  char *old_msg;
 
   /* create the zwrite and send the message */
-  owl_zwrite_create_from_line(&z, line);
-  owl_zwrite_populate_zsig(&z);
+  owl_zwrite_populate_zsig(z);
   if (msg) {
-    owl_zwrite_set_message(&z, msg);
+    owl_zwrite_set_message(z, msg);
   }
-
-  mymsg=owl_zwrite_get_message(&z);
+  old_msg = owl_strdup(owl_zwrite_get_message(z));
 
   zcrypt = owl_sprintf("%s/zcrypt", owl_get_bindir());
   argv[0] = "zcrypt";
   argv[1] = "-E";
-  argv[2] = "-c"; argv[3] = owl_zwrite_get_class(&z);
-  argv[4] = "-i"; argv[5] = owl_zwrite_get_instance(&z);
+  argv[2] = "-c"; argv[3] = owl_zwrite_get_class(z);
+  argv[4] = "-i"; argv[5] = owl_zwrite_get_instance(z);
   argv[6] = NULL;
 
-  rv = call_filter(zcrypt, argv, mymsg, &cryptmsg, &status);
+  rv = call_filter(zcrypt, argv, owl_zwrite_get_message(z), &cryptmsg, &status);
 
   owl_free(zcrypt);
 
   if (rv || status) {
     if(cryptmsg) owl_free(cryptmsg);
+    owl_free(old_msg);
     owl_function_error("Error in zcrypt, possibly no key found.  Message not sent.");
     owl_function_beep();
-    owl_zwrite_cleanup(&z);
     return;
   }
 
-  owl_zwrite_set_message(&z, cryptmsg);
-  owl_zwrite_set_opcode(&z, "crypt");
-    
-  owl_zwrite_send_message(&z);
+  owl_zwrite_set_message_raw(z, cryptmsg);
+  owl_zwrite_set_opcode(z, "crypt");
+
+  owl_zwrite_send_message(z);
   owl_function_makemsg("Waiting for ack...");
 
   /* If it's personal */
-  if (owl_zwrite_is_personal(&z)) {
-    /* create the outgoing message */
-    mymsg=owl_zwrite_get_message(&z);
-    m=owl_function_make_outgoing_zephyr(mymsg, line, owl_zwrite_get_zsig(&z));
+  if (owl_zwrite_is_personal(z)) {
+    /* Create the outgoing message. Restore the un-crypted message for display. */
+    owl_zwrite_set_message_raw(z, old_msg);
+    m=owl_function_make_outgoing_zephyr(z);
     if (m) {
       owl_global_messagequeue_addmsg(&g, m);
     } else {
@@ -451,7 +473,6 @@ void owl_function_zcrypt(const char *line, const char *msg)
 
   /* free the zwrite */
   owl_free(cryptmsg);
-  owl_zwrite_cleanup(&z);
 }
 
 void owl_callback_aimwrite(owl_editwin *e) {
@@ -807,8 +828,6 @@ void owl_function_lastmsg(void)
 void owl_function_shift_right(void)
 {
   owl_global_set_rightshift(&g, owl_global_get_rightshift(&g)+10);
-  owl_mainwin_redisplay(owl_global_get_mainwin(&g));
-  owl_global_set_needrefresh(&g);
 }
 
 void owl_function_shift_left(void)
@@ -818,8 +837,6 @@ void owl_function_shift_left(void)
   shift=owl_global_get_rightshift(&g);
   if (shift > 0) {
     owl_global_set_rightshift(&g, MAX(shift - 10, 0));
-    owl_mainwin_redisplay(owl_global_get_mainwin(&g));
-    owl_global_set_needrefresh(&g);
   } else {
     owl_function_beep();
     owl_function_makemsg("Already full left");
@@ -981,23 +998,14 @@ void owl_function_quit(void)
   if (owl_global_get_newmsgproc_pid(&g)) {
     kill(owl_global_get_newmsgproc_pid(&g), SIGHUP);
   }
-
-  /* Quit zephyr */
-  owl_zephyr_shutdown();
   
   /* Quit AIM */
   if (owl_global_is_aimloggedin(&g)) {
     owl_aim_logout();
   }
 
-  /* done with curses */
-  endwin();
-
-  /* restore terminal settings */
-  tcsetattr(0, TCSAFLUSH, owl_global_get_startup_tio(&g));
-
   owl_function_debugmsg("Quitting Owl");
-  exit(0);
+  owl_select_quit_loop();
 }
 
 void owl_function_calculate_topmsg(int direction)
@@ -1203,7 +1211,7 @@ void owl_function_debugmsg(const char *fmt, ...)
   if (!owl_global_is_debug_fast(&g))
     return;
 
-  file = fopen(owl_global_get_debug_file(&g), "a");
+  file = owl_global_get_debug_file_handle(&g);
   if (!file) /* XXX should report this */
     return;
 
@@ -1213,7 +1221,7 @@ void owl_function_debugmsg(const char *fmt, ...)
 	  (int) getpid(), ctime(&now), now - owl_global_get_starttime(&g));
   vfprintf(file, fmt, ap);
   putc('\n', file);
-  fclose(file);
+  fflush(file);
 
   va_end(ap);
 }
@@ -1222,7 +1230,6 @@ void owl_function_beep(void)
 {
   if (owl_global_is_bell(&g)) {
     beep();
-    owl_global_set_needrefresh(&g); /* do we really need this? */
   }
 }
 
@@ -1251,30 +1258,19 @@ void owl_function_unsubscribe(const char *class, const char *inst, const char *r
   }
 }
 
-void owl_function_set_cursor(WINDOW *win)
-{
-  /* Be careful that this window is actually empty, otherwise panels get confused */
-  if (is_wintouched(win)) {
-    owl_function_debugmsg("Warning: owl_function_set_cursor called on dirty window");
-    update_panels();
-  }
-  wnoutrefresh(win);
+static void _dirty_everything(owl_window *w) {
+  if (!owl_window_is_realized(w))
+    return;
+  owl_window_dirty(w);
+  owl_window_children_foreach(w, (GFunc)_dirty_everything, NULL);
 }
 
 void owl_function_full_redisplay(void)
 {
-  redrawwin(owl_global_get_curs_recwin(&g));
-  redrawwin(owl_global_get_curs_sepwin(&g));
-  /* Work around curses segfualts with windows off the screen */
-  if (g.lines >= owl_global_get_typwin_lines(&g)+2)
-      redrawwin(owl_global_get_curs_typwin(&g));
-  if (g.lines >= 2)
-      redrawwin(owl_global_get_curs_msgwin(&g));
-
-  sepbar("");
-  owl_function_makemsg("");
-
-  owl_global_set_needrefresh(&g);
+  /* Ask every widget to redraw itself. */
+  _dirty_everything(owl_window_get_screen());
+  /* Force ncurses to redisplay everything. */
+  clearok(stdscr, TRUE);
 }
 
 void owl_function_popless_text(const char *text)
@@ -1282,16 +1278,18 @@ void owl_function_popless_text(const char *text)
   owl_popwin *pw;
   owl_viewwin *v;
 
-  pw=owl_global_get_popwin(&g);
-  v=owl_global_get_viewwin(&g);
-
+  if (owl_global_get_popwin(&g) || owl_global_get_viewwin(&g)) {
+    owl_function_error("Popwin already in use.");
+    return;
+  }
+  pw = owl_popwin_new();
+  owl_global_set_popwin(&g, pw);
   owl_popwin_up(pw);
-  owl_global_push_context(&g, OWL_CTX_POPLESS, v, "popless");
-  owl_viewwin_init_text(v, owl_popwin_get_curswin(pw),
-			owl_popwin_get_lines(pw), owl_popwin_get_cols(pw),
-			text);
-  owl_viewwin_redisplay(v);
-  owl_global_set_needrefresh(&g);
+
+  v = owl_viewwin_new_text(owl_popwin_get_content(pw), text);
+  owl_global_set_viewwin(&g, v);
+
+  owl_global_push_context(&g, OWL_CTX_POPLESS, v, "popless", NULL);
 }
 
 void owl_function_popless_fmtext(const owl_fmtext *fm)
@@ -1299,16 +1297,18 @@ void owl_function_popless_fmtext(const owl_fmtext *fm)
   owl_popwin *pw;
   owl_viewwin *v;
 
-  pw=owl_global_get_popwin(&g);
-  v=owl_global_get_viewwin(&g);
-
+  if (owl_global_get_popwin(&g) || owl_global_get_viewwin(&g)) {
+    owl_function_error("Popwin already in use.");
+    return;
+  }
+  pw = owl_popwin_new();
+  owl_global_set_popwin(&g, pw);
   owl_popwin_up(pw);
-  owl_global_push_context(&g, OWL_CTX_POPLESS, v, "popless");
-  owl_viewwin_init_fmtext(v, owl_popwin_get_curswin(pw),
-		   owl_popwin_get_lines(pw), owl_popwin_get_cols(pw),
-		   fm);
-  owl_viewwin_redisplay(v);
-  owl_global_set_needrefresh(&g);
+
+  v = owl_viewwin_new_fmtext(owl_popwin_get_content(pw), fm);
+  owl_global_set_viewwin(&g, v);
+
+  owl_global_push_context(&g, OWL_CTX_POPLESS, v, "popless", NULL);
 }
 
 void owl_function_popless_file(const char *filename)
@@ -1568,13 +1568,12 @@ void owl_function_page_curmsg(int step)
   
   /* redisplay */
   owl_mainwin_redisplay(owl_global_get_mainwin(&g));
-  owl_global_set_needrefresh(&g);
 }
 
 void owl_function_resize_typwin(int newsize)
 {
   owl_global_set_typwin_lines(&g, newsize);
-  owl_function_resize();
+  owl_mainpanel_layout_contents(&g.mainpanel);
 }
 
 void owl_function_mainwin_pagedown(void)
@@ -1637,7 +1636,7 @@ void owl_function_printallvars(void)
   owl_variable_dict_namelist_cleanup(&varnames);
 
   owl_function_popless_text(str->str);
-  g_string_free(str, TRUE);
+  g_string_free(str, true);
 }
 
 void owl_function_show_variables(void)
@@ -1688,7 +1687,6 @@ void owl_function_delete_by_id(int id, int flag)
       owl_message_unmark_delete(m);
     }
     owl_mainwin_redisplay(owl_global_get_mainwin(&g));
-    owl_global_set_needrefresh(&g);
   } else {
     owl_function_error("No message with id %d: unable to mark for (un)delete",id);
   }
@@ -1724,7 +1722,6 @@ void owl_function_delete_automsgs(void)
   }
   owl_mainwin_redisplay(owl_global_get_mainwin(&g));
   owl_function_makemsg("%i messages marked for deletion", count);
-  owl_global_set_needrefresh(&g);
 }
 
 void owl_function_status(void)
@@ -1925,48 +1922,49 @@ void owl_callback_command(owl_editwin *e)
 void owl_function_start_command(const char *line)
 {
   owl_editwin *tw;
+  owl_context *ctx;
 
   tw = owl_global_set_typwin_active(&g, OWL_EDITWIN_STYLE_ONELINE, owl_global_get_cmd_history(&g));
 
   owl_editwin_set_locktext(tw, "command: ");
-  owl_global_set_needrefresh(&g);
 
   owl_editwin_insert_string(tw, line);
-  owl_editwin_redisplay(tw);
 
-  owl_global_push_context(&g, OWL_CTX_EDITLINE, tw, "editline");
+  ctx = owl_editcontext_new(OWL_CTX_EDITLINE, tw, "editline",
+                            owl_global_deactivate_editcontext, &g);
+  owl_global_push_context_obj(&g, ctx);
   owl_editwin_set_callback(tw, owl_callback_command);
 }
 
 owl_editwin *owl_function_start_question(const char *line)
 {
   owl_editwin *tw;
+  owl_context *ctx;
 
   tw = owl_global_set_typwin_active(&g, OWL_EDITWIN_STYLE_ONELINE, owl_global_get_cmd_history(&g));
 
   owl_editwin_set_locktext(tw, line);
-  owl_global_set_needrefresh(&g);
 
-  owl_editwin_redisplay(tw);
-
-  owl_global_push_context(&g, OWL_CTX_EDITRESPONSE, tw, "editresponse");
+  ctx = owl_editcontext_new(OWL_CTX_EDITRESPONSE, tw, "editresponse",
+                            owl_global_deactivate_editcontext, &g);
+  owl_global_push_context_obj(&g, ctx);
   return tw;
 }
 
 owl_editwin *owl_function_start_password(const char *line)
 {
   owl_editwin *tw;
+  owl_context *ctx;
 
   tw = owl_global_set_typwin_active(&g, OWL_EDITWIN_STYLE_ONELINE, NULL);
 
   owl_editwin_set_echochar(tw, '*');
 
   owl_editwin_set_locktext(tw, line);
-  owl_global_set_needrefresh(&g);
 
-  owl_editwin_redisplay(tw);
-
-  owl_global_push_context(&g, OWL_CTX_EDITRESPONSE, tw, "editresponse");
+  ctx = owl_editcontext_new(OWL_CTX_EDITRESPONSE, tw, "editresponse",
+                            owl_global_deactivate_editcontext, &g);
+  owl_global_push_context_obj(&g, ctx);
   return tw;
 }
 
@@ -1994,22 +1992,18 @@ char *owl_function_exec(int argc, const char *const *argv, const char *buff, int
   buff = skiptokens(buff, 1);
   newbuff = owl_sprintf("%s%s", buff, redirect);
 
-  if (type == 1) {
+  if (type == OWL_OUTPUT_POPUP) {
     owl_popexec_new(newbuff);
   } else {
     p = popen(newbuff, "r");
     out = owl_slurp(p);
     pclose(p);
     
-    if (type==1) {
-      owl_function_popless_text(out);
-    } else if (type==0) {
+    if (type == OWL_OUTPUT_RETURN) {
       owl_free(newbuff);
       return out;
-    } else if (type==2) {
+    } else if (type == OWL_OUTPUT_ADMINMSG) {
       owl_function_adminmsg(buff, out);
-    } else {
-      owl_function_popless_text(out);
     }
     owl_free(out);
   }
@@ -2036,14 +2030,12 @@ char *owl_function_perl(int argc, const char *const *argv, const char *buff, int
 
   perlout = owl_perlconfig_execute(buff);
   if (perlout) { 
-    if (type==1) {
+    if (type == OWL_OUTPUT_POPUP) {
       owl_function_popless_text(perlout);
-    } else if (type==2) {
+    } else if (type == OWL_OUTPUT_ADMINMSG) {
       owl_function_adminmsg(buff, perlout);
-    } else if (type==0) {
+    } else if (type == OWL_OUTPUT_RETURN) {
       return perlout;
-    } else {
-      owl_function_popless_text(perlout);
     }
     owl_free(perlout);
   }
@@ -2135,7 +2127,6 @@ void owl_function_create_filter(int argc, const char *const *argv)
       return;
     }
     owl_filter_set_fgcolor(f, owl_util_string_to_color(argv[3]));
-    owl_global_set_needrefresh(&g);
     owl_mainwin_redisplay(owl_global_get_mainwin(&g));
     return;
   }
@@ -2150,7 +2141,6 @@ void owl_function_create_filter(int argc, const char *const *argv)
       return;
     }
     owl_filter_set_bgcolor(f, owl_util_string_to_color(argv[3]));
-    owl_global_set_needrefresh(&g);
     owl_mainwin_redisplay(owl_global_get_mainwin(&g));
     return;
   }
@@ -2179,7 +2169,6 @@ void owl_function_create_filter(int argc, const char *const *argv)
   if (inuse) {
     owl_function_change_currentview_filter(argv[1]);
   }
-  owl_global_set_needrefresh(&g);
   owl_mainwin_redisplay(owl_global_get_mainwin(&g));
 }
 
@@ -2290,15 +2279,16 @@ void owl_function_show_zpunts(void)
 
 /* Create a filter for a class, instance if one doesn't exist.  If
  * instance is NULL then catch all messgaes in the class.  Returns the
- * name of the filter, which the caller must free.
+ * name of the filter or null.  The caller must free this name.
  * If 'related' is nonzero, encompass unclasses and .d classes as well.
  */
 char *owl_function_classinstfilt(const char *c, const char *i, int related) 
 {
   owl_filter *f;
-  char *argbuff, *filtname;
+  char *filtname;
   char *tmpclass, *tmpinstance = NULL;
   char *class, *instance = NULL;
+  GString *buf;
 
   if (related) {
     class = owl_util_baseclass(c);
@@ -2338,31 +2328,38 @@ char *owl_function_classinstfilt(const char *c, const char *i, int related)
 
   /* create the new filter */
   tmpclass=owl_text_quote(class, OWL_REGEX_QUOTECHARS, OWL_REGEX_QUOTEWITH);
-  owl_text_tr(tmpclass, ' ', '.');
-  owl_text_tr(tmpclass, '\'', '.');
-  owl_text_tr(tmpclass, '"', '.');
   if (instance) {
     tmpinstance=owl_text_quote(instance, OWL_REGEX_QUOTECHARS, OWL_REGEX_QUOTEWITH);
-    owl_text_tr(tmpinstance, ' ', '.');
-    owl_text_tr(tmpinstance, '\'', '.');
-    owl_text_tr(tmpinstance, '"', '.');
   }
 
-  argbuff = owl_sprintf(related ? "class ^(un)*%s(\\.d)*$" : "class ^%s$", tmpclass);
+  buf = g_string_new("");
+  owl_string_appendf_quoted(buf,
+                            related ? "class ^(un)*%q(\\.d)*$" : "class ^%q$",
+                            tmpclass);
+
   if (tmpinstance) {
-    char *tmp = argbuff;
-    argbuff = owl_sprintf(related ? "%s and ( instance ^(un)*%s(\\.d)*$ )" : "%s and instance ^%s$", tmp, tmpinstance);
-    owl_free(tmp);
+    owl_string_appendf_quoted(buf,
+                              related ?
+                              " and ( instance ^(un)*%q(\\.d)*$ )" :
+                              " and instance ^%q$",
+                              tmpinstance);
   }
   owl_free(tmpclass);
-  if (tmpinstance) owl_free(tmpinstance);
+  owl_free(tmpinstance);
 
-  f = owl_filter_new_fromstring(filtname, argbuff);
+  f = owl_filter_new_fromstring(filtname, buf->str);
+  g_string_free(buf, true);
+  if (f == NULL) {
+    /* Couldn't make a filter for some reason. Return NULL. */
+    owl_function_error("Error creating filter '%s'", filtname);
+    owl_free(filtname);
+    filtname = NULL;
+    goto done;
+  }
 
   /* add it to the global list */
   owl_global_add_filter(&g, f);
 
-  owl_free(argbuff);
 done:
   owl_free(class);
   if (instance) {
@@ -2373,22 +2370,20 @@ done:
 
 /* Create a filter for personal zephyrs to or from the specified
  * zephyr user.  Includes login/logout notifications for the user.
- * The name of the filter will be 'user-<user>'.  If a filter already
+ * The name of the filter will be 'user-<shortuser>'.  If a filter already
  * exists with this name, no new filter will be created.  This allows
  * the configuration to override this function.  Returns the name of
  * the filter, which the caller must free.
  */
-char *owl_function_zuserfilt(const char *user)
+char *owl_function_zuserfilt(const char *longuser)
 {
   owl_filter *f;
-  char *argbuff, *longuser, *esclonguser, *shortuser, *filtname;
-
-  /* stick the local realm on if it's not there */
-  longuser=long_zuser(user);
-  shortuser=short_zuser(user);
+  char *argbuff, *esclonguser, *shortuser, *filtname;
 
   /* name for the filter */
-  filtname=owl_sprintf("user-%s", shortuser);
+  shortuser = short_zuser(longuser);
+  filtname = owl_sprintf("user-%s", shortuser);
+  owl_free(shortuser);
 
   /* if it already exists then go with it.  This lets users override */
   if (owl_global_get_filter(&g, filtname)) {
@@ -2398,22 +2393,25 @@ char *owl_function_zuserfilt(const char *user)
   /* create the new-internal filter */
   esclonguser = owl_text_quote(longuser, OWL_REGEX_QUOTECHARS, OWL_REGEX_QUOTEWITH);
 
-  argbuff=owl_sprintf("( type ^zephyr$ and filter personal and "
-      "( ( direction ^in$ and sender ^%1$s$ ) or ( direction ^out$ and "
-      "recipient ^%1$s$ ) ) ) or ( ( class ^login$ ) and ( sender ^%1$s$ ) )",
-      esclonguser);
+  argbuff=owl_string_build_quoted("( type ^zephyr$ and filter personal and "
+      "( ( direction ^in$ and sender ^%q$ ) or ( direction ^out$ and "
+      "recipient ^%q$ ) ) ) or ( ( class ^login$ ) and ( sender ^%q$ ) )",
+      esclonguser, esclonguser, esclonguser);
+  owl_free(esclonguser);
 
   f = owl_filter_new_fromstring(filtname, argbuff);
+  owl_free(argbuff);
+
+  if (f == NULL) {
+    /* Couldn't make a filter for some reason. Return NULL. */
+    owl_function_error("Error creating filter '%s'", filtname);
+    owl_free(filtname);
+    return NULL;
+  }
 
   /* add it to the global list */
   owl_global_add_filter(&g, f);
 
-  /* free stuff */
-  owl_free(argbuff);
-  owl_free(longuser);
-  owl_free(esclonguser);
-  owl_free(shortuser);
-
   return(filtname);
 }
 
@@ -2444,16 +2442,20 @@ char *owl_function_aimuserfilt(const char *user)
       "( type ^aim$ and ( ( sender ^%1$s$ and recipient ^%2$s$ ) or "
       "( sender ^%2$s$ and recipient ^%1$s$ ) ) )",
       escuser, owl_global_get_aim_screenname_for_filters(&g));
+  owl_free(escuser);
 
   f = owl_filter_new_fromstring(filtname, argbuff);
+  owl_free(argbuff);
+
+  if (f == NULL) {
+    owl_function_error("Error creating filter '%s'", filtname);
+    owl_free(filtname);
+    return NULL;
+  }
 
   /* add it to the global list */
   owl_global_add_filter(&g, f);
 
-  /* free stuff */
-  owl_free(argbuff);
-  owl_free(escuser);
-
   return(filtname);
 }
 
@@ -2473,17 +2475,21 @@ char *owl_function_typefilt(const char *type)
   /* create the new-internal filter */
   esctype = owl_text_quote(type, OWL_REGEX_QUOTECHARS, OWL_REGEX_QUOTEWITH);
 
-  argbuff = owl_sprintf("type ^%s$", esctype);
+  argbuff = owl_string_build_quoted("type ^%q$", esctype);
+  owl_free(esctype);
 
   f = owl_filter_new_fromstring(filtname, argbuff);
+  owl_free(argbuff);
+
+  if (f == NULL) {
+    owl_function_error("Error creating filter '%s'", filtname);
+    owl_free(filtname);
+    return NULL;
+  }
 
   /* add it to the global list */
   owl_global_add_filter(&g, f);
 
-  /* free stuff */
-  owl_free(argbuff);
-  owl_free(esctype);
-
   return filtname;
 }
 
@@ -2509,6 +2515,43 @@ void owl_function_delete_curview_msgs(int flag)
   owl_mainwin_redisplay(owl_global_get_mainwin(&g));  
 }
 
+static char *owl_function_smartfilter_cc(const owl_message *m) {
+  const char *ccs;
+  char *ccs_quoted;
+  char *filtname;
+  owl_filter *f;
+  GString *buf;
+
+  ccs = owl_message_get_attribute_value(m, "zephyr_ccs");
+
+  filtname = owl_sprintf("conversation-%s", ccs);
+  owl_text_tr(filtname, ' ', '-');
+
+  if (owl_global_get_filter(&g, filtname)) {
+    return filtname;
+  }
+
+  buf = g_string_new("type ^zephyr$ and filter personal and "
+                     "zephyr_ccs ^");
+  ccs_quoted = owl_text_quote(ccs, OWL_REGEX_QUOTECHARS, OWL_REGEX_QUOTEWITH);
+  owl_string_append_quoted_arg(buf, ccs_quoted);
+  g_string_append_c(buf, '$');
+  owl_free(ccs_quoted);
+
+  f = owl_filter_new_fromstring(filtname, buf->str);
+  g_string_free(buf, true);
+
+  if (f == NULL) {
+    owl_function_error("Error creating filter '%s'", filtname);
+    owl_free(filtname);
+    return NULL;
+  }
+
+  owl_global_add_filter(&g, f);
+
+  return filtname;
+}
+
 /* Create a filter based on the current message.  Returns the name of
  * a filter or null.  The caller must free this name.
  *
@@ -2527,8 +2570,8 @@ char *owl_function_smartfilter(int type, int invert_related)
 {
   const owl_view *v;
   const owl_message *m;
-  char *zperson, *filtname=NULL;
-  const char *argv[2];
+  char *filtname = NULL;
+  const char *argv[2], *zperson;
   int related = owl_global_is_narrow_related(&g) ^ invert_related;
 
   v=owl_global_get_current_view(&g);
@@ -2562,14 +2605,17 @@ char *owl_function_smartfilter(int type, int invert_related)
   /* narrow personal and login messages to the sender or recip as appropriate */
   if (owl_message_is_type_zephyr(m)) {
     if (owl_message_is_personal(m) || owl_message_is_loginout(m)) {
-      if (owl_message_is_direction_in(m)) {
-        zperson=short_zuser(owl_message_get_sender(m));
-      } else {
-        zperson=short_zuser(owl_message_get_recipient(m));
+      if (owl_message_get_attribute_value(m, "zephyr_ccs") != NULL) {
+        return owl_function_smartfilter_cc(m);
       }
-      filtname=owl_function_zuserfilt(zperson);
-      owl_free(zperson);
-      return(filtname);
+
+      if (owl_message_is_direction_in(m)) {
+        zperson = owl_message_get_sender(m);
+      } else {
+        zperson = owl_message_get_recipient(m);
+      }
+      filtname = owl_function_zuserfilt(zperson);
+      return filtname;
     }
 
     /* narrow class MESSAGE, instance foo, recip * messages to class, inst */
@@ -2599,8 +2645,8 @@ void owl_function_smartzpunt(int type)
    * If type=0, uses just class.  If type=1, uses instance as well. */
   const owl_view *v;
   const owl_message *m;
-  const char *cmdprefix, *mclass, *minst;
-  char *cmd;
+  const char *mclass, *minst;
+  GString *buf;
   
   v=owl_global_get_current_view(&g);
   m=owl_view_get_element(v, owl_global_get_curmsg(&g));
@@ -2626,22 +2672,16 @@ void owl_function_smartzpunt(int type)
     owl_function_error("smartzpunt can't safely do this for <%s,%s>",
 			 mclass, minst);
   } else {
-    cmdprefix = "start-command zpunt ";
-    cmd = owl_malloc(strlen(cmdprefix)+strlen(mclass)+strlen(minst)+10);
-    strcpy(cmd, cmdprefix);
-    strcat(cmd, owl_getquoting(mclass));
-    strcat(cmd, mclass);
-    strcat(cmd, owl_getquoting(mclass));
+    buf = g_string_new("start-command zpunt ");
+    owl_string_append_quoted_arg(buf, mclass);
     if (type) {
-      strcat(cmd, " ");
-      strcat(cmd, owl_getquoting(minst));
-      strcat(cmd, minst);
-      strcat(cmd, owl_getquoting(minst));
+      g_string_append_c(buf, ' ');
+      owl_string_append_quoted_arg(buf, minst);
     } else {
-      strcat(cmd, " *");
+      g_string_append(buf, " *");
     }
-    owl_function_command(cmd);
-    owl_free(cmd);
+    owl_function_command(buf->str);
+    g_string_free(buf, true);
   }
 }
 
@@ -2691,7 +2731,6 @@ int owl_function_color_filter(const char *filtname, const char *fgcolor, const c
   }
   owl_filter_set_fgcolor(f, owl_util_string_to_color(fgcolor));
   
-  owl_global_set_needrefresh(&g);
   owl_mainwin_redisplay(owl_global_get_mainwin(&g));
   return(0);
 }
@@ -2745,49 +2784,37 @@ void owl_function_show_colors(void)
  */
 void owl_function_zpunt(const char *class, const char *inst, const char *recip, int direction)
 {
-  char *puntexpr, *classexpr, *instexpr, *recipexpr;
+  GString *buf;
   char *quoted;
 
+  buf = g_string_new("");
   if (!strcmp(class, "*")) {
-    classexpr = owl_sprintf("class .*");
+    g_string_append(buf, "class .*");
   } else {
     quoted=owl_text_quote(class, OWL_REGEX_QUOTECHARS, OWL_REGEX_QUOTEWITH);
-    owl_text_tr(quoted, ' ', '.');
-    owl_text_tr(quoted, '\'', '.');
-    owl_text_tr(quoted, '"', '.');
-    classexpr = owl_sprintf("class ^(un)*%s(\\.d)*$", quoted);
+    owl_string_appendf_quoted(buf, "class ^(un)*%q(\\.d)*$", quoted);
     owl_free(quoted);
   }
   if (!strcmp(inst, "*")) {
-    instexpr = owl_sprintf(" and instance .*");
+    g_string_append(buf, " and instance .*");
   } else {
     quoted=owl_text_quote(inst, OWL_REGEX_QUOTECHARS, OWL_REGEX_QUOTEWITH);
-    owl_text_tr(quoted, ' ', '.');
-    owl_text_tr(quoted, '\'', '.');
-    owl_text_tr(quoted, '"', '.');
-    instexpr = owl_sprintf(" and instance ^(un)*%s(\\.d)*$", quoted);
+    owl_string_appendf_quoted(buf, " and instance ^(un)*%q(\\.d)*$", quoted);
     owl_free(quoted);
   }
   if (!strcmp(recip, "*")) {
-    recipexpr = owl_sprintf("");
+    /* g_string_append(buf, ""); */
   } else {
     if(!strcmp(recip, "%me%")) {
       recip = owl_zephyr_get_sender();
     }
     quoted=owl_text_quote(recip, OWL_REGEX_QUOTECHARS, OWL_REGEX_QUOTEWITH);
-    owl_text_tr(quoted, ' ', '.');
-    owl_text_tr(quoted, '\'', '.');
-    owl_text_tr(quoted, '"', '.');
-    recipexpr = owl_sprintf(" and recipient ^%s$", quoted);
+    owl_string_appendf_quoted(buf, " and recipient ^%q$", quoted);
     owl_free(quoted);
   }
 
-  puntexpr = owl_sprintf("%s %s %s", classexpr, instexpr, recipexpr);
-  owl_function_punt(puntexpr, direction);
-  owl_free(puntexpr);
-  owl_free(classexpr);
-  owl_free(instexpr);
-  owl_free(recipexpr);
+  owl_function_punt(buf->str, direction);
+  g_string_free(buf, true);
 }
 
 void owl_function_punt(const char *filter, int direction)
@@ -2855,7 +2882,7 @@ void owl_function_show_keymaps(void)
     kmname = owl_list_get_element(&l, i);
     km = owl_keyhandler_get_keymap(kh, kmname);
     owl_fmtext_append_bold(&fm, "\n\n----------------------------------------------------------------------------------------------------\n\n");
-    owl_keymap_get_details(km, &fm);    
+    owl_keymap_get_details(km, &fm, 0);
   }
   owl_fmtext_append_normal(&fm, "\n");
   
@@ -2881,7 +2908,7 @@ void owl_function_show_keymap(const char *name)
   owl_fmtext_init_null(&fm);
   km = owl_keyhandler_get_keymap(owl_global_get_keyhandler(&g), name);
   if (km) {
-    owl_keymap_get_details(km, &fm);
+    owl_keymap_get_details(km, &fm, 1);
   } else {
     owl_fmtext_append_normal(&fm, "No such keymap...\n");
   }  
@@ -2899,10 +2926,8 @@ void owl_function_help_for_command(const char *cmdname)
   owl_fmtext_cleanup(&fm);
 }
 
-void owl_function_search_start(const char *string, int direction)
+void owl_function_set_search(const char *string)
 {
-  /* direction is OWL_DIRECTION_DOWNWARDS or OWL_DIRECTION_UPWARDS or
-   * OWL_DIRECTION_NONE */
   owl_regex re;
 
   if (string && owl_regex_create_quoted(&re, string) == 0) {
@@ -2911,27 +2936,16 @@ void owl_function_search_start(const char *string, int direction)
   } else {
     owl_global_set_search_re(&g, NULL);
   }
-
-  if (direction == OWL_DIRECTION_NONE)
-    owl_mainwin_redisplay(owl_global_get_mainwin(&g));
-  else
-    owl_function_search_helper(0, direction);
 }
 
-void owl_function_search_continue(int direction)
-{
-  /* direction is OWL_DIRECTION_DOWNWARDS or OWL_DIRECTION_UPWARDS */
-  owl_function_search_helper(1, direction);
-}
-
-void owl_function_search_helper(int mode, int direction)
+void owl_function_search_helper(int consider_current, int direction)
 {
   /* move to a message that contains the string.  If direction is
    * OWL_DIRECTION_DOWNWARDS then search fowards, if direction is
    * OWL_DIRECTION_UPWARDS then search backwards.
    *
-   * If mode==0 then it will stay on the current message if it
-   * contains the string.
+   * If consider_current is true then it will stay on the
+   * current message if it contains the string.
    */
 
   const owl_view *v;
@@ -2947,7 +2961,7 @@ void owl_function_search_helper(int mode, int direction)
     return;
   }
 
-  if (mode==0) {
+  if (consider_current) {
     start=curmsg;
   } else if (direction==OWL_DIRECTION_DOWNWARDS) {
     start=curmsg+1;
@@ -3405,18 +3419,14 @@ void owl_function_showerrs(void)
 void owl_function_makemsg(const char *fmt, ...)
 {
   va_list ap;
-  char buff[2048];
-
-  if (!owl_global_get_curs_msgwin(&g)) return;
+  char *str;
 
   va_start(ap, fmt);
-  werase(owl_global_get_curs_msgwin(&g));
-  
-  vsnprintf(buff, 2048, fmt, ap);
-  owl_function_debugmsg("makemsg: %s", buff);
-  waddstr(owl_global_get_curs_msgwin(&g), buff);  
-  owl_global_set_needrefresh(&g);
+  str = g_strdup_vprintf(fmt, ap);
   va_end(ap);
+
+  owl_function_debugmsg("makemsg: %s", str);
+  owl_msgwin_set_text_nocopy(&g.msgwin, str);
 }
 
 /* get locations for everyone in .anyone.  If 'notify' is '1' then
@@ -3429,7 +3439,6 @@ void owl_function_zephyr_buddy_check(int notify)
 #ifdef HAVE_LIBZEPHYR
   int i, j;
   owl_list anyone;
-  owl_zbuddylist *zbl;
   GList **zaldlist;
   GList *zaldptr;
   ZAsyncLocateData_t *zald;
@@ -3437,7 +3446,6 @@ void owl_function_zephyr_buddy_check(int notify)
 
   if (!owl_global_is_havezephyr(&g)) return;
   owl_global_set_pseudologin_notify(&g, notify);
-  zbl = owl_global_get_zephyr_buddylist(&g);
   zaldlist = owl_global_get_zaldlist(&g);
 
   /* Clear the existing ZALDs first. */
